import json
import os
from typing import Any, Dict, List

import pytest
from attr import define

from featureflags.evaluations.auth_target import Target
from featureflags.evaluations.evaluator import Evaluator
from featureflags.evaluations.feature import FeatureConfig, FeatureConfigKind
from featureflags.evaluations.segment import Segment
from featureflags.lru_cache import LRUCache
from featureflags.repository import Repository

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
PATH = cwd = BASE_PATH + "/ff-test-cases/tests/"

cache = LRUCache()
repository = Repository(cache)
evaluator = Evaluator(repository)


@define
class Test:
    flag: str
    target: str
    expected: Any

    @classmethod
    def from_dict(cls, source: Dict[str, Any]) -> 'Test':
        return Test(source["flag"], source.get("target", "default"),
                    source["expected"])


@define
class Feature:
    flags: List[FeatureConfig]
    segments: List[Segment]
    targets: List[Target]
    tests: List[Test]

    @classmethod
    def from_dict(cls, source: Dict[str, Any]) -> 'Feature':
        flags = [FeatureConfig.from_dict(flag) for flag in source["flags"]]

        segments = []
        if "segments" in source:
            segments = [Segment.from_dict(segment)
                        for segment in source["segments"]]
        targets = []
        if "targets" in source:
            targets = [Target.from_dict(target) for
                       target in source["targets"]]

        tests = [Test.from_dict(test) for test in source["tests"]]

        return Feature(flags=flags, segments=segments, targets=targets,
                       tests=tests)


@define
class TestCase:
    name: str
    flag: str
    target: str
    expected: str
    targets: List[Target]


def load_json_test_file(filename: str) -> Feature:
    result = None
    with open(filename, 'r') as f:
        result = Feature.from_dict(json.load(f))
    return result


def load_test_files():
    result = []
    for r, d, f in os.walk(PATH):
        for file in f:
            if '.json' in file:
                fname = os.path.join(r, file)
                print(f'Loading File: {fname}.')
                usecase = load_json_test_file(fname)
                # usecase.flag.feature += file
                result.append((fname, usecase))
    return result


def usecase_provider():
    usecases = load_test_files()

    result = []
    for fname, usecase in usecases:
        for flag in usecase.flags:
            repository.set_flag(flag)

        for segment in usecase.segments:
            repository.set_segment(segment)

        for test_case in usecase.tests:
            result.append(
                TestCase(name='given flag {} then target {} should get {}'
                         .format(test_case.flag,
                                 test_case.target,
                                 test_case.expected),
                         flag=test_case.flag,
                         target=test_case.target,
                         expected=test_case.expected,
                         targets=usecase.targets))

    return result


def feature_name(tc):
    return tc.name


@pytest.mark.parametrize('tc', usecase_provider(), ids=feature_name)
def test_evaluator(tc: TestCase):
    target = None
    if tc.target != "_no_target":
        target = next(
            (val for val in tc.targets if val.identifier == tc.target),
            None
        )

    got = None

    switch_kind = {
        FeatureConfigKind.BOOLEAN:
            lambda: evaluator.evaluate(tc.flag, target)
            .bool(target, tc.flag, default=False),
        FeatureConfigKind.STRING:
            lambda: evaluator.evaluate(tc.flag, target).string(
                target, tc.flag,
                default="failed"),
        FeatureConfigKind.INT:
            lambda: evaluator.evaluate(tc.flag, target)
            .number(target, tc.flag, default=0.100),
        FeatureConfigKind.JSON:
            lambda: evaluator.evaluate(tc.flag, target)
            .json(target, tc.flag, default={}),
    }

    kind = repository.get_flag(tc.flag).kind
    got = switch_kind[kind]()

    if kind == FeatureConfigKind.JSON:
        tc.expected = json.loads(tc.expected)

    assert got == tc.expected



import time
from threading import Event, Thread

from featureflags.api.client import AuthenticatedClient
from featureflags.repository import DataProviderInterface

from .api.default.get_all_segments import sync as retrieve_segments
from .api.default.get_feature_config import sync as retrieve_flags
from .config import Config
from .sdk_logging_codes import (info_poll_started, info_polling_stopped,
                                info_sdk_init_ok, info_sdk_init_waiting)
from .util import log


class PollingProcessor(Thread):

    def __init__(self, client: AuthenticatedClient, config: Config,
                 environment_id: str, wait_for_initialization: Event,
                 ready: Event, stream_ready: Event,
                 repository: DataProviderInterface) -> None:
        Thread.__init__(self)
        self.daemon = True
        self.__environment_id = environment_id
        self.__client = client
        self.__config = config
        self.__running = False
        self.__wait_for_initialization = wait_for_initialization
        self.__ready = ready
        self.__stream_ready = stream_ready
        self.__repository = repository

    def run(self):
        if not self.__running:
            if self.__config.pull_interval < 60:
                log.warning("Pull Interval must be greater than or equal to "
                            "60 seconds, was: " +
                            str(self.__config.pull_interval) +
                            " setting to 60")
                self.__config.pull_interval = 60

            self.__running = True
            #  Get initial flags and groups
            try:
                info_sdk_init_waiting()
                log.info("Fetching initial target segments and flags")
                self.retrieve_flags_and_segments()
                log.info("Initial target segments and flags fetched. "
                         "PollingProcessor will start in: " +
                         str(self.__config.pull_interval) + " seconds")
                #  Segments and flags have been cached so
                #  mark the Client as initialised.
                self.__wait_for_initialization.set()
                info_sdk_init_ok()
            except Exception as ex:
                log.exception(
                    'Error: Exception encountered when '
                    'getting initial flags and segments. %s',
                    ex
                )
            #  Sleep for an interval before going into the polling loop.
            time.sleep(self.__config.pull_interval)
            info_poll_started(self.__config.pull_interval)
            while self.__running:
                start_time = time.time()
                try:
                    if self.__config.enable_stream and \
                            self.__stream_ready.is_set():
                        info_polling_stopped('streaming mode is active')
                        # on stream disconnect, make sure flags are in sync
                        self.retrieve_flags_and_segments()
                        #  Block until ready.set() is called
                        self.__ready.wait()
                        log.debug('Poller resuming ')
                    else:
                        self.retrieve_flags_and_segments()
                        self.__ready.set()
                except Exception as e:
                    log.exception(
                        'Error: Exception encountered when polling flags. %s',
                        e
                    )

                elapsed = time.time() - start_time
                if elapsed < self.__config.pull_interval:
                    log.info("Poller sleeping for " +
                             (self.__config.pull_interval - elapsed).__str__())
                    " seconds"
                    time.sleep(self.__config.pull_interval - elapsed)

    def stop(self):
        self.__running = False
        info_polling_stopped("Client was closed")

    def retrieve_flags_and_segments(self):
        t1 = Thread(target=self.__retrieve_segments)
        t2 = Thread(target=self.__retrieve_flags)
        t1.start()
        t2.start()
        t1.join()
        t2.join()

    def __retrieve_flags(self):
        log.debug("Loading feature flags")
        flags = retrieve_flags(
            client=self.__client, environment_uuid=self.__environment_id
        )
        log.debug("Feature flags loaded")
        for flag in flags:
            log.debug("Put flag %s into repository", flag.feature)
            self.__repository.set_flag(flag)

    def __retrieve_segments(self):
        log.debug("Loading target segments")
        segments = retrieve_segments(
            client=self.__client, environment_uuid=self.__environment_id
        )
        log.debug("Target segments loaded")
        for segment in segments:
            log.debug("Put %s segment into repository", segment.identifier)
            self.__repository.set_segment(segment)



import json

import pytest

from featureflags.evaluations.auth_target import Target
from featureflags.evaluations.variation import Variation


@pytest.mark.parametrize(
    "test_input, target, flag_identifier, default,expected",
    [
        ("true", Target(identifier='harness'), "flag1", False, True),
        ("false", Target(identifier='harness'), "flag1", False, False),
        ("", Target(identifier='harness'), "flag1", False, False),
    ],
)
def test_variation_bool(test_input, target, flag_identifier,
                        default, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.bool(target, flag_identifier, default=default)

    assert got == expected


@pytest.mark.parametrize(
    "test_input, target, flag_identifier, default,expected",
    [
        ("test", Target(identifier='harness'), "flag1", "test_default",
         "test"),
        (None, Target(identifier='harness'), "flag1", "test_default",
         "test_default"),
    ],
)
def test_variation_string(test_input, target, flag_identifier,
                          default, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.string(target, flag_identifier, default=default)

    assert got == expected


@pytest.mark.parametrize(
    "test_input,target,flag_identifier,default,expected", [
        (1.1, Target(identifier='harness'), "flag1", 1.0, 1.1),
        (None, Target(identifier='harness'), "flag1", 1.0, 1.0)
    ]
)
def test_variation_number(test_input, target, flag_identifier,
                          default, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.number(target, flag_identifier, default=default)

    assert got == expected


@pytest.mark.parametrize(
    "test_input,target,flag_identifier,expected",
    [
        (json.dumps({"field": "value"}), Target(identifier='harness'),
         "flag1", json.dumps({"field": "value"})),
        (None, Target(identifier='harness'), "flag1",
         json.dumps({"field": "value2"})),
    ],
)
def test_variation_json(test_input, target,
                        flag_identifier, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.json(target, flag_identifier, default={"field": "value2"})

    assert got == json.loads(expected)



import json
from typing import Any, Dict, List, Type, TypeVar, Union
from xmlrpc.client import boolean

import attr

from featureflags.evaluations.auth_target import Target
from featureflags.models import UNSET, Unset
from featureflags.util import log

T = TypeVar("T", bound="Variation")


@attr.s(auto_attribs=True)
class Variation(object):
    identifier: str
    value: Union[str, None]
    name: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def bool(self, target: Target, flag_identifier: str,
             default: bool = False) -> bool:
        if self.value:
            result = self.value.lower() == "true"
            log.debug(
                "SDKCODE:6000: Evaluated bool variation successfully:"
                "%s", {"result": result, "flag identifier": flag_identifier,
                       "target": target})
            return result
        log.error(
            "SDKCODE:6001: Failed to evaluate bool variation for %s and the "
            "default variation '%s' is being returned",
            {"target": target, "flag": flag_identifier}, default)
        return default

    def string(self, target: Target, flag_identifier: str,
               default: str) -> str:
        if self.value:
            result = self.value
            log.debug(
                "SDKCODE:6000: Evaluated string variation successfully:"
                "%s", {"result": result, "flag identifier": flag_identifier,
                       "target": target})
            return result
        log.error(
            "SDKCODE:6001: Failed to evaluate string variation for %s and the"
            " default variation '%s' is being returned",
            {"target": target, "flag": flag_identifier}, default)
        return default

    def number(self, target: Target, flag_identifier: str,
               default: float) -> float:
        if self.value:
            result = float(self.value)
            log.debug(
                "SDKCODE:6000: Evaluated number variation successfully:"
                "%s", {"result": result, "flag identifier": flag_identifier,
                       "target": target})
            return result
        log.error(
            "SDKCODE:6001: Failed to evaluate number variation for %s and the"
            " default variation '%s' is being returned",
            {"target": target, "flag": flag_identifier}, default)
        return default

    def int(self, target: Target, flag_identifier: str,
            default: int) -> int:
        if self.value:
            result = int(self.value)
            log.debug(
                "SDKCODE:6000: Evaluated number variation successfully:"
                "%s", {"result": result, "flag identifier": flag_identifier,
                       "target": target})
            return result
        log.error(
            "SDKCODE:6001: Failed to evaluate int variation for %s and the "
            "default variation '%s' is being returned",
            {"target": target, "flag": flag_identifier}, default)
        return default

    def json(self, target: Target, flag_identifier: str,
             default: dict) -> dict:
        if self.value:
            result = json.loads(self.value)
            log.debug(
                "SDKCODE:6000: Evaluated json variation successfully:"
                "%s", {"result": result, "flag identifier": flag_identifier,
                       "target": target})
            return result
        log.error(
            "SDKCODE:6001: Failed to evaluate json variation for %s and the "
            "default variation '%s' is being returned",
            {"target": target, "flag": flag_identifier}, default)
        return default

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier
        value = self.value
        name = self.name
        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "value": value,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")

        value = d.pop("value")

        name = d.pop("name", UNSET)

        description = d.pop("description", UNSET)

        variation = cls(
            identifier=identifier,
            value=value,
            name=name,
            description=description,
        )

        variation.additional_properties = d
        return variation

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> boolean:
        return key in self.additional_properties



import random
import threading
import time
from threading import Thread
from typing import Union

from featureflags.repository import DataProviderInterface

from .api.client import AuthenticatedClient
from .api.default.get_feature_config_by_identifier import \
    sync as get_feature_config
from .api.default.get_segment_by_identifier import sync as get_target_segment
from .config import Config
from .models.message import Message
from .sdk_logging_codes import (info_stream_connected,
                                info_stream_event_received,
                                info_stream_stopped, warn_stream_disconnected,
                                warn_stream_retrying)
from .sse_client import SSEClient
from .util import log

BACK_OFF_IN_SECONDS = 5


class StreamProcessor(Thread):
    def __init__(self, repository: DataProviderInterface,
                 client: AuthenticatedClient,
                 environment_id: str, api_key: str, token: str,
                 config: Config,
                 ready: threading.Event,
                 poller: threading.Event,
                 cluster: str):

        Thread.__init__(self)
        self.daemon = True
        self._running = False
        self._ready = ready
        self.poller = poller
        self._client = client
        self._environment_id = environment_id
        self._api_key = api_key
        self._token = token
        self._stream_url = f'{config.base_url}/stream?cluster={cluster}'
        self._repository = repository

    def run(self):
        log.info("Starting StreamingProcessor connecting to uri: " +
                 self._stream_url)
        self._running = True
        retries = 0
        while self._running:
            try:
                messages = self._connect()
                info_stream_connected()
                self.poller.clear()  # were streaming now, so tell any poller
                # threads calling wait to wait...
                self._ready.set()
                retries = 0  # reset the retry counter
                for msg in messages:
                    if not self._running:
                        break
                    if msg.data:
                        info_stream_event_received(msg.data)
                        message = Message.from_str(msg.data)
                        self.process_message(message)
                    if self._ready.is_set() is False:
                        self._ready.set()
            except Exception as e:
                warn_stream_disconnected(e)
                # Signal the poller than it should start due to stream error.
                if self.poller.is_set() is False:
                    self.poller.set()

                # Calculate back of sleep
                sleep = (BACK_OFF_IN_SECONDS * 2 ** retries +
                         random.uniform(0, 1))
                warn_stream_retrying(f'{sleep.__str__()}s')
                time.sleep(sleep)
                retries += 1

    def _connect(self) -> SSEClient:
        return SSEClient(self._stream_url, headers={
            'Authorization': f'Bearer {self._token}',
            'API-Key': self._api_key
        })

    def process_message(self, msg: Message) -> None:
        processor: Union[FlagMsgProcessor, SegmentMsgProcessor, None] = None
        if msg.domain == "flag":
            log.debug('Starting flag message processor with %s', msg)
            processor = FlagMsgProcessor(repository=self._repository,
                                         client=self._client,
                                         environment_id=self._environment_id,
                                         msg=msg)
        elif msg.domain == "target-segment":
            log.debug('Starting segment message processor with %s', msg)
            processor = SegmentMsgProcessor(
                repository=self._repository,
                client=self._client,
                environment_id=self._environment_id,
                msg=msg
            )
        if processor:
            processor.start()

    def stop(self):
        self._running = False
        info_stream_stopped()


class FlagMsgProcessor(Thread):

    def __init__(self, repository: DataProviderInterface,
                 client: AuthenticatedClient,
                 environment_id: str, msg: Message):
        Thread.__init__(self)
        self._repository = repository
        self._client = client
        self._environemnt_id = environment_id
        self._msg = msg

    def run(self):
        log.debug("Fetching flag config '%s' from server",
                  self._msg.identifier)
        fc = get_feature_config(client=self._client,
                                identifier=self._msg.identifier,
                                environment_uuid=self._environemnt_id)
        log.info("Feature config '%s' loaded", fc.feature)
        if self._msg.event == 'create' or self._msg.event == 'patch':
            self._repository.set_flag(fc)
            log.info('flag %s successfully stored in the cache', fc.feature)
        elif self._msg.event == 'delete':
            self._repository.remove_flag(fc)
            log.info('flag %s successfully removed from cache', fc.feature)


class SegmentMsgProcessor(Thread):

    def __init__(self, repository: DataProviderInterface,
                 client: AuthenticatedClient,
                 environment_id: str, msg: Message):
        Thread.__init__(self)
        self._repository = repository
        self._client = client
        self._environemnt_id = environment_id
        self._msg = msg

    def run(self):
        log.debug("Fetching target segment '%s' from server",
                  self._msg.identifier)
        ts = get_target_segment(client=self._client,
                                identifier=self._msg.identifier,
                                environment_uuid=self._environemnt_id)
        log.info("Target segment '%s' loaded", ts.identifier)
        if self._msg.event == 'create' or self._msg.event == 'patch':
            self._repository.set_segment(ts)
            log.info('flag %s successfully stored in cache', ts.identifier)
        elif self._msg.event == 'delete':
            self._repository.remove_segment(ts.identifier)
            log.info('flag %s successfully removed from cache', ts.identifier)



from featureflags.ftypes.boolean import Boolean


def test_equal(mocker):
    input = [True, False]

    m = mocker.patch("featureflags.ftypes.boolean.get_value",
                     return_value=True)
    boolean = Boolean(value=True)

    got = boolean.equal(input)
    expected = True

    assert got == expected
    m.assert_called_with(input)



import featureflags.sdk_logging_codes as sdk_codes
from featureflags.evaluations.auth_target import Target


def test_logs_dont_raise_exception():
    target = Target(identifier='harness', name="asd")
    sdk_codes.info_poll_started(60)
    sdk_codes.info_sdk_init_ok()
    sdk_codes.info_sdk_init_waiting()
    sdk_codes.info_sdk_start_close()
    sdk_codes.info_sdk_close_success()
    sdk_codes.info_sdk_auth_ok()
    sdk_codes.info_polling_stopped("streaming active")
    sdk_codes.info_stream_connected()
    sdk_codes.info_stream_event_received("")
    sdk_codes.info_stream_stopped()
    sdk_codes.info_metrics_thread_started(10)
    sdk_codes.info_metrics_thread_existed()
    sdk_codes.info_metrics_success()
    sdk_codes.info_metrics_target_exceeded()
    sdk_codes.warn_auth_failed_srv_defaults()
    sdk_codes.warn_failed_init_auth_error()
    sdk_codes.warn_auth_failed_exceed_retries()
    sdk_codes.wan_missing_sdk_key()
    sdk_codes.warn_auth_retying(1, "some error")
    sdk_codes.warn_stream_disconnected("example reason")
    sdk_codes.warn_stream_retrying(5)
    sdk_codes.warn_post_metrics_failed("example reason")
    sdk_codes.warn_post_metrics_target_batch_failed("example reason")
    sdk_codes.warn_default_variation_served("identifier", target, "default")



"""Configuration is a base class that has default values that you can change
during the instance of the client class"""

from typing import Callable

from .interface import Cache
from .lru_cache import LRUCache
from .util import log

BASE_URL = "https://config.ff.harness.io/api/1.0"
EVENTS_URL = "https://events.ff.harness.io/api/1.0"
MINUTE = 60
PULL_INTERVAL = 1 * MINUTE
PERSIST_INTERVAL = 1 * MINUTE
EVENTS_SYNC_INTERVAL = 1 * MINUTE


class Config(object):
    def __init__(
            self,
            base_url: str = BASE_URL,
            events_url: str = EVENTS_URL,
            pull_interval: int = PULL_INTERVAL,
            persist_interval: int = PERSIST_INTERVAL,
            events_sync_interval: int = EVENTS_SYNC_INTERVAL,
            cache: Cache = None,
            store: object = None,
            enable_stream: bool = True,
            enable_analytics: bool = True,
            max_auth_retries: int = 10
    ):
        self.base_url = base_url
        self.events_url = events_url
        self.pull_interval = pull_interval
        self.persist_interval = persist_interval
        if events_sync_interval < EVENTS_SYNC_INTERVAL:
            log.warning("Metrics events sync interval cannot be lower than "
                        "60 seconds. Default of 60 seconds will be used")
            self.events_sync_interval = EVENTS_SYNC_INTERVAL
        else:
            self.events_sync_interval = events_sync_interval

        self.cache = cache
        if self.cache is None:
            self.cache = LRUCache()
        self.store = store
        self.enable_stream = enable_stream
        self.enable_analytics = enable_analytics
        self.max_auth_retries = max_auth_retries


default_config = Config()


def with_base_url(base_url: str) -> Callable:
    def func(config: Config) -> None:
        config.base_url = base_url

    return func


def with_events_url(events_url: str) -> Callable:
    def func(config: Config) -> None:
        config.events_url = events_url

    return func


def with_stream_enabled(value: bool) -> Callable:
    def func(config: Config) -> None:
        config.enable_stream = value

    return func


def with_analytics_enabled(value: bool) -> Callable:
    def func(config: Config) -> None:
        config.enable_analytics = value

    return func


def with_pull_interval(value: int) -> Callable:
    def func(config: Config) -> None:
        config.pull_interval = value

    return func


def with_max_auth_retries(value: int) -> Callable:
    def func(config: Config) -> None:
        config.max_auth_retries = value

    return func



from typing import Any, Dict, Optional, Type, TypeVar, Union

import attr

from featureflags.ftypes import TYPES
from featureflags.ftypes.interface import Interface
from featureflags.models import UNSET, Unset
from featureflags.util import log

T = TypeVar("T", bound="Target")


@attr.s(auto_attribs=True)
class Target():

    identifier: str
    name: Union[Unset, str] = UNSET
    anonymous: Union[Unset, bool] = UNSET
    attributes: Union[Unset, Dict[str, Any]] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier
        name = self.name
        anonymous = self.anonymous

        attributes: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "identifier": identifier,
            }
        )

        if name is not UNSET:
            field_dict["name"] = name
        if anonymous is not UNSET:
            field_dict["anonymous"] = anonymous
        if attributes is not UNSET:
            field_dict["attributes"] = attributes

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")
        name = d.pop("name", UNSET)
        anonymous = d.pop("anonymous", UNSET)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, Dict[str, Any]]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = _attributes

        target = cls(
            identifier=identifier,
            name=name,
            anonymous=anonymous,
            attributes=attributes,
        )

        return target

    def get_attr_value(self, attribute: str) -> Optional[str]:
        if not attribute:
            log.debug("Attribute is empty")
            return None
        result: Any = getattr(self, attribute, None)
        if not result and not isinstance(self.attributes,
                                         Unset):
            log.debug("Checking attributes field %s", attribute)
            result = self.attributes.get(attribute, None)
            if result is None:
                log.warning("Attribute %s does not exist", attribute)
        log.debug("Target %s attribute %s value %s", self, attribute, result)
        return result

    def get_type(self, attribute: str) -> Optional[Interface]:
        value: Optional[str] = self.get_attr_value(attribute)
        for _type, klass in TYPES.items():
            if isinstance(value, _type):
                operator = TYPES.get(_type, None)
                if operator:
                    return klass(value)
        log.debug("Unsupported type found on attribute %s", attribute)
        return None



import pytest

from featureflags.ftypes.utils import (get_float_value, get_int_value,
                                       get_str_value, get_value)


@pytest.mark.parametrize(
    ["input", "expected"],
    [
        (True, True),
        ([True, False], True),
        ([True, True], True),
        ([False, True], False),
        ([False, False], False),
    ],
)
def test_get_value(input, expected):
    got = get_value(input)

    assert got == expected


def test_get_str_value(mocker):
    value = "value"
    m = mocker.patch("featureflags.ftypes.utils.get_value", return_value=value)

    val = get_str_value([value])

    m.assert_called_with([value])
    assert val == value


def test_get_int_value(mocker):
    value = 1
    m = mocker.patch("featureflags.ftypes.utils.get_value", return_value=value)

    val = get_int_value([value])

    m.assert_called_with([value])
    assert val == value


def test_get_number_value(mocker):
    value = 1.1
    m = mocker.patch("featureflags.ftypes.utils.get_value", return_value=value)

    val = get_float_value([value])

    m.assert_called_with([value])
    assert val == value



from featureflags.util import log


def get_sdk_code_message(key):
    sdk_codes = {
        # SDK_INIT_1xxx
        1000: "The SDK has successfully initialized",
        1001: "The SDK has failed to initialize due to an authentication "
              "error - defaults will be served",
        1002: "The SDK has failed to initialize due to a missing or empty "
              "API key - defaults will be served",
        1003: "The SDK is waiting for initialization to complete",
        # SDK_AUTH_2xxx
        2000: "Authentication was successful",
        2001: "Authentication failed with a non-recoverable error",
        2002: "Authentication attempt",
        2003: "Authentication failed and max retries have been exceeded",
        # SDK_CLOSE_3xxx
        3000: "Closing SDK",
        3001: "SDK Closed successfully",
        # SDK_POLL_4xxx
        4000: "Polling started, intervalMs:",
        4001: "Polling stopped, reason:",
        # SDK_STREAM_5xxx
        5000: "SSE stream successfully connected",
        5001: "SSE stream disconnected, reason:",
        5002: "SSE event received: ",
        5003: "SSE retrying to connect in",
        5004: "SSE stopped",
        # SDK_EVAL_6xxx - these are hardcoded in `variation.py` as they
        # are more complex
        # SDK_METRICS_7xxx
        7000: "Metrics thread started with request interval:",
        7001: "Metrics thread exited",
        7002: "Posting metrics failed, reason:",
        7003: "Metrics posted successfully",
        7004: "Target metrics exceeded max size, remaining targets for this "
              "analytics interval will not be sent",
        7005: "Target metrics batches succeeded:",
        7006: "Target metrics batch/batches failed:",
    }
    if key in sdk_codes:
        return sdk_codes[key]
    else:
        return "Unknown SDK code"


def sdk_err_msg(error_code, append_text=""):
    return f"SDKCODE:{error_code}: {get_sdk_code_message(error_code)} " \
           f"{append_text} "


def wan_missing_sdk_key():
    msg = sdk_err_msg(1002)
    log.warning(msg)


def info_poll_started(duration_sec):
    log.info(sdk_err_msg(4000, duration_sec * 1000))


def info_sdk_init_ok():
    log.info(sdk_err_msg(1000))


def info_sdk_init_waiting():
    log.info(sdk_err_msg(1003))


def info_sdk_start_close():
    log.info(sdk_err_msg(3000))


def info_sdk_close_success():
    log.info(sdk_err_msg(3001))


def info_sdk_auth_ok():
    log.info(sdk_err_msg(2000))


def info_polling_stopped(reason):
    log.info(sdk_err_msg(4001, reason))


def info_stream_connected():
    log.info(sdk_err_msg(5000))


def info_stream_event_received(event_json):
    log.info(sdk_err_msg(5002, event_json))


def info_stream_stopped():
    log.info(sdk_err_msg(5004))


def info_metrics_thread_started(interval):
    log.info(sdk_err_msg(7000, interval))


def info_metrics_success():
    log.info(sdk_err_msg(7003))


def info_metrics_target_exceeded():
    log.info(sdk_err_msg(7004))


def info_metrics_target_batch_success(message):
    log.info(sdk_err_msg(7005, message))


def info_metrics_thread_existed():
    log.info(sdk_err_msg(7001))


def info_eval_success():
    log.info(sdk_err_msg(6000))


def warn_auth_failed_srv_defaults():
    log.warning(sdk_err_msg(2001))


def warn_failed_init_auth_error():
    log.warning(sdk_err_msg(1001))


def warn_auth_failed_exceed_retries():
    log.warning(sdk_err_msg(2003))


def warn_auth_retying(attempt, error):
    log.warning(sdk_err_msg(2002,
                            f"attempt {attempt}, got error: {error}, "
                            f"Retrying ..."))


def warn_stream_disconnected(reason):
    log.warning(sdk_err_msg(5001, reason))


def warn_stream_retrying(seconds):
    log.warning(sdk_err_msg(5003, seconds))


def warn_post_metrics_failed(reason):
    log.warning(sdk_err_msg(7002, reason))


def warn_post_metrics_target_batch_failed(message):
    log.warning(sdk_err_msg(7006, message))


def warn_default_variation_served(flag, target, default):
    log.warning(sdk_err_msg(6001,
                            f"flag={flag}, "
                            f"target={target}, default={default}"))



from typing import Any, Dict, Optional, Union

import httpx
from tenacity import (Retrying, retry_all, retry_if_result, stop_after_attempt,
                      wait_exponential)

from featureflags.api.client import Client
from featureflags.api.types import Response
from featureflags.models.authentication_request import AuthenticationRequest
from featureflags.models.authentication_response import AuthenticationResponse
from featureflags.sdk_logging_codes import warn_auth_retying
from featureflags.util import log


class UnrecoverableAuthenticationException(Exception):
    pass


def _get_kwargs(
        *,
        client: Client,
        json_body: AuthenticationRequest,
) -> Dict[str, Any]:
    url = "{}/client/auth".format(client.base_url)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    json_json_body = json_body.to_dict()

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "json": json_json_body,
    }


def _parse_response(
        *, response: httpx.Response
) -> Optional[Union[AuthenticationResponse, None]]:
    if response.status_code == 200:
        return AuthenticationResponse.from_dict(response.json())
    else:
        raise UnrecoverableAuthenticationException(
            f'Authentication failed on an unrecoverable error: {response}')


def _build_response(
        *, response: httpx.Response
) -> Response[Union[AuthenticationResponse, None]]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
        *,
        client: Client,
        json_body: AuthenticationRequest,
) -> Response[Union[AuthenticationResponse, None]]:
    kwargs = _get_kwargs(
        client=client,
        json_body=json_body,
    )
    max_auth_retries = client.get_max_auth_retries()
    response = _post_request(kwargs, max_auth_retries)
    return _build_response(response=response)


def handle_http_result(response):
    code = response.status_code
    if code in {408, 425, 429, 500, 502, 503, 504}:
        return True
    else:
        log.error(
            f'Authentication received HTTP code #{code} and '
            'will not attempt to reconnect')
        return False


def _post_request(kwargs, max_auth_retries):
    retryer = Retrying(
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_all(
            retry_if_result(lambda response: response.status_code != 200),
            retry_if_result(handle_http_result)
        ),
        before_sleep=lambda retry_state: warn_auth_retying(
            retry_state.attempt_number,
            retry_state.outcome.result()),
        stop=stop_after_attempt(max_auth_retries),
    )
    return retryer(httpx.post, **kwargs)


def sync(
        *,
        client: Client,
        json_body: AuthenticationRequest,
) -> Optional[Union[AuthenticationResponse, None]]:
    """Used to retrieve all target segments for certain account id."""

    return sync_detailed(
        client=client,
        json_body=json_body,
    ).parsed


async def asyncio_detailed(
        *,
        client: Client,
        json_body: AuthenticationRequest,
) -> Response[Union[AuthenticationResponse, None]]:
    kwargs = _get_kwargs(
        client=client,
        json_body=json_body,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.post(**kwargs)

    return _build_response(response=response)


async def asyncio(
        *,
        client: Client,
        json_body: AuthenticationRequest,
) -> Optional[Union[AuthenticationResponse, None]]:
    """Used to retrieve all target segments for certain account id."""

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
        )
    ).parsed



import typing

import attr

from featureflags.ftypes.interface import Interface

from .utils import get_int_value


@attr.s(auto_attribs=True)
class Integer(Interface):

    value: int

    def starts_with(self, value: typing.Any) -> bool:
        return False

    def ends_with(self, value: typing.Any) -> bool:
        return False

    def match(self, value: typing.Any) -> bool:
        return False

    def contains(self, value: typing.Any) -> bool:
        return False

    def equal_sensitive(self, value: typing.Any) -> bool:
        return False

    def equal(self, value: typing.Any) -> bool:
        _value = get_int_value(value)
        return self.value == _value

    def greater_than(self, value: typing.Any) -> bool:
        _value = get_int_value(value)
        return self.value > _value

    def greater_than_equal(self, value: typing.Any) -> bool:
        _value = get_int_value(value)
        return self.value >= _value

    def less_than(self, value: typing.Any) -> bool:
        _value = get_int_value(value)
        return self.value < _value

    def less_than_equal(self, value: typing.Any) -> bool:
        _value = get_int_value(value)
        return self.value <= _value

    def in_list(self, value: typing.List[typing.Any]) -> bool:
        return self.value in value



""" Contains all the data models used in inputs/outputs """

from .authentication_request import AuthenticationRequest
from .authentication_request_target import AuthenticationRequestTarget
from .authentication_request_target_attributes import \
    AuthenticationRequestTargetAttributes
from .authentication_response import AuthenticationResponse
from .metrics import Metrics
from .metrics_data import MetricsData
from .metrics_data_metrics_type import MetricsDataMetricsType
from .target_data import TargetData
from .unset import UNSET, Unset

__all__ = [
    'AuthenticationResponse',
    'AuthenticationRequestTargetAttributes',
    'AuthenticationRequestTarget',
    'AuthenticationRequest',

    'Metrics',
    'MetricsData',
    'MetricsDataMetricsType',
    'TargetData',

    'UNSET',
    'Unset',
]



import abc
import typing


class Interface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "starts_with")
            and callable(subclass.starts_with)
            and hasattr(subclass, "ends_with")
            and callable(subclass.ends_with)
            and hasattr(subclass, "match")
            and callable(subclass.match)
            and hasattr(subclass, "contains")
            and callable(subclass.contains)
            and hasattr(subclass, "equal_sensitive")
            and callable(subclass.equal_sensitive)
            and hasattr(subclass, "equal")
            and callable(subclass.equal)
            and hasattr(subclass, "greater_than")
            and callable(subclass.greater_than)
            and hasattr(subclass, "greater_than_equal")
            and callable(subclass.greater_than_equal)
            and hasattr(subclass, "less_than")
            and callable(subclass.less_than)
            and hasattr(subclass, "less_than_equal")
            and callable(subclass.less_than_equal)
            and hasattr(subclass, "in_list")
            and callable(subclass.in_list)
            or NotImplemented
        )

    @abc.abstractmethod
    def starts_with(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def ends_with(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def match(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def contains(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def equal_sensitive(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def equal(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def greater_than(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def greater_than_equal(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def less_than(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def less_than_equal(self, value: typing.Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def in_list(self, value: typing.List[typing.Any]) -> bool:
        raise NotImplementedError



"""Client for interacting with Harness FF server"""

import threading
from typing import Any, Callable, Dict, Optional

from jwt import decode
from tenacity import RetryError

import featureflags.sdk_logging_codes as sdk_codes
from featureflags.analytics import AnalyticsService
from featureflags.evaluations.evaluator import Evaluator
from featureflags.repository import Repository

from .api.client import AuthenticatedClient, Client
from .api.default.authenticate import (AuthenticationRequest,
                                       UnrecoverableAuthenticationException)
from .api.default.authenticate import sync as authenticate
from .config import Config, default_config
from .evaluations.auth_target import Target
from .polling import PollingProcessor
from .streaming import StreamProcessor

VERSION: str = "1.0"


class MissingOrEmptyAPIKeyException(Exception):
    pass


class CfClient(object):
    def __init__(
            self, sdk_key: str,
            *options: Callable,
            config: Optional[Config] = None
    ):
        self._client: Optional[Client] = None
        #  The Client is considered initialized when flags and groups
        #  are loaded into cache.
        self._initialized = threading.Event()
        # Keep track if initialization has failed due to authentication
        # or a missing/empty API key.
        self._initialized_failed = False
        self._auth_token: Optional[str] = None
        self._environment_id: Optional[str] = None
        self._sdk_key: Optional[str] = sdk_key
        self._config: Config = default_config
        self._cluster: str = '1'

        if config:
            self._config = config

        for option in options:
            if callable(option):
                option(self._config)

        if self._config.cache is None:
            raise Exception("cache cannot be none")

        self._repository = Repository(self._config.cache)
        self._evaluator = Evaluator(self._repository)

        self.run()

    def run(self):
        try:
            if self._sdk_key is None or self._sdk_key == "":
                raise MissingOrEmptyAPIKeyException()
            self.authenticate()
            sdk_codes.info_sdk_auth_ok()
            streaming_event = threading.Event()
            polling_event = threading.Event()

            self._polling_processor = PollingProcessor(
                client=self._client,
                config=self._config,
                environment_id=self._environment_id,
                #  PollingProcessor is responsible for doing the initial
                #  flag/group fetch and cache. So we allocate it the
                #  responsibility
                #  for setting the Client is_initialized variable.
                wait_for_initialization=self._initialized,
                ready=polling_event,
                stream_ready=streaming_event,
                repository=self._repository
            )
            self._polling_processor.start()

            if self._config.enable_stream:
                self._stream = StreamProcessor(
                    repository=self._repository,
                    client=self._client,
                    environment_id=self._environment_id,
                    api_key=self._sdk_key,
                    token=self._auth_token,
                    config=self._config,
                    ready=streaming_event,
                    poller=polling_event,
                    cluster=self._cluster,
                )
                self._stream.start()

            if self._config.enable_analytics:
                self._analytics = AnalyticsService(
                    config=self._config,
                    client=self._client,
                    environment=self._environment_id
                )

        except RetryError:
            sdk_codes.warn_auth_failed_exceed_retries()
            sdk_codes.warn_failed_init_auth_error()
            self._initialized_failed = True
            # We just need to unblock the thread here
            # in case wait_for_intialization was called. The SDK has already
            # logged that authentication failed and defaults will be returned.
            self._initialized.set()
        except UnrecoverableAuthenticationException:
            self._initialized_failed = True
            sdk_codes.warn_auth_failed_srv_defaults()
            sdk_codes.warn_failed_init_auth_error()
            # Same again, unblock the thread.
            self._initialized.set()
        except MissingOrEmptyAPIKeyException:
            self._initialized_failed = True
            sdk_codes.wan_missing_sdk_key()
            # And again, unblock the thread.
            self._initialized.set()

    def wait_for_initialization(self):
        self._initialized.wait()

    def is_initialized(self):
        if self._initialized_failed:
            return False
        return self._initialized.is_set()

    def get_environment_id(self):
        return self._environment_id

    def authenticate(self):
        client = Client(
            base_url=self._config.base_url,
            events_url=self._config.events_url,
            max_auth_retries=self._config.max_auth_retries
        )
        body = AuthenticationRequest(api_key=self._sdk_key)
        response = authenticate(client=client, json_body=body)
        self._auth_token = response.auth_token

        decoded = decode(self._auth_token, options={
            "verify_signature": False})
        self._environment_id = decoded["environment"]
        self._cluster = decoded["clusterIdentifier"]
        if not self._cluster:
            self._cluster = '1'
        self._client = AuthenticatedClient(
            base_url=self._config.base_url,
            events_url=self._config.events_url,
            token=self._auth_token,
            params={
                'cluster': self._cluster
            },
            max_auth_retries=self._config.max_auth_retries
        )
        # Additional headers used to track usage
        additional_headers = {
            "User-Agent": "PythonSDK/" + VERSION,
            "Harness-SDK-Info": f'Python {VERSION} Server',
            "Harness-EnvironmentID": self._environment_id,
            "Harness-AccountID": decoded["accountID"]}
        self._client = self._client.with_headers(additional_headers)

    def bool_variation(self, identifier: str, target: Target,
                       default: bool) -> bool:
        variation = self._evaluator.evaluate(identifier, target)
        # Only register metrics if analytics is enabled,
        # and sometimes when the SDK starts up we can
        # evaluate before the flag is cached which results in
        # an empty identifier.
        if self._config.enable_analytics and variation.identifier != "":
            self._analytics.enqueue(target, identifier, variation)
        return variation.bool(target, identifier, default)

    def int_variation(self, identifier: str, target: Target,
                      default: int) -> int:
        variation = self._evaluator.evaluate(identifier, target)
        # Only register metrics if analytics is enabled,
        # and sometimes when the SDK starts up we can
        # evaluate before the flag is cached which results in
        # an empty identifier.
        if self._config.enable_analytics and variation.identifier != "":
            self._analytics.enqueue(target, identifier, variation)
        return variation.int(target, identifier, default)

    def number_variation(self, identifier: str, target: Target,
                         default: float) -> float:
        variation = self._evaluator.evaluate(
            identifier, target)
        # Only register metrics if analytics is enabled,
        # and sometimes when the SDK starts up we can
        # evaluate before the flag is cached which results in
        # an empty identifier.
        if self._config.enable_analytics and variation.identifier != "":
            self._analytics.enqueue(target, identifier, variation)
        return variation.number(target, identifier, default)

    def string_variation(self, identifier: str, target: Target,
                         default: str) -> str:
        variation = self._evaluator.evaluate(
            identifier, target)
        # Only register metrics if analytics is enabled,
        # and sometimes when the SDK starts up we can
        # evaluate before the flag is cached which results in
        # an empty identifier.
        if self._config.enable_analytics and variation.identifier != "":
            self._analytics.enqueue(target, identifier, variation)
        return variation.string(target, identifier, default)

    def json_variation(self, identifier: str, target: Target,
                       default: Dict[str, Any]) -> Dict[str, Any]:
        variation = self._evaluator.evaluate(identifier, target)
        # Only register metrics if analytics is enabled,
        # and sometimes when the SDK starts up we can
        # evaluate before the flag is cached which results in
        # an empty identifier.
        if self._config.enable_analytics and variation.identifier != "":
            self._analytics.enqueue(target, identifier, variation)
        return variation.json(target, identifier, default)

    def close(self):
        sdk_codes.info_sdk_start_close()
        self._polling_processor.stop()
        if self._config.enable_stream:
            self._stream.stop()

        if self._config.enable_analytics:
            self._analytics.close()
        sdk_codes.info_sdk_close_success()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.authentication_request_target_attributes import \
    AuthenticationRequestTargetAttributes
from .unset import UNSET, Unset

T = TypeVar("T", bound="AuthenticationRequestTarget")


@attr.s(auto_attribs=True)
class AuthenticationRequestTarget:
    """ """

    identifier: str
    name: Union[Unset, str] = UNSET
    anonymous: Union[Unset, bool] = UNSET
    attributes: Union[Unset, AuthenticationRequestTargetAttributes] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier
        name = self.name
        anonymous = self.anonymous
        attributes: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if anonymous is not UNSET:
            field_dict["anonymous"] = anonymous
        if attributes is not UNSET:
            field_dict["attributes"] = attributes

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")

        name = d.pop("name", UNSET)

        anonymous = d.pop("anonymous", UNSET)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, AuthenticationRequestTargetAttributes]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = AuthenticationRequestTargetAttributes.from_dict(
                _attributes
            )

        authentication_request_target = cls(
            identifier=identifier,
            name=name,
            anonymous=anonymous,
            attributes=attributes,
        )

        authentication_request_target.additional_properties = d
        return authentication_request_target

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from featureflags.models import UNSET, Unset

from .target_map import TargetMap

T = TypeVar("T", bound="VariationMap")


@attr.s(auto_attribs=True)
class VariationMap:
    """ """

    variation: str
    targets: Union[Unset, List[TargetMap]] = UNSET
    target_segments: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        variation = self.variation
        targets: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.targets, Unset):
            targets = []
            for targets_item_data in self.targets:
                targets_item = targets_item_data.to_dict()

                targets.append(targets_item)

        target_segments: Union[Unset, List[str]] = UNSET
        if not isinstance(self.target_segments, Unset):
            target_segments = self.target_segments

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "variation": variation,
            }
        )
        if targets is not UNSET:
            field_dict["targets"] = targets
        if target_segments is not UNSET:
            field_dict["targetSegments"] = target_segments

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        variation = d.pop("variation")

        targets = []
        _targets = d.pop("targets", UNSET)
        for targets_item_data in _targets or []:
            targets_item = TargetMap.from_dict(targets_item_data)

            targets.append(targets_item)

        target_segments = cast(List[str], d.pop("targetSegments", UNSET))

        variation_map = cls(
            variation=variation,
            targets=targets,
            target_segments=target_segments,
        )

        variation_map.additional_properties = d
        return variation_map

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



ONE_HUNDRED: int = 100

SEGMENT_MATCH_OPERATOR = "segmentMatch"
IN_OPERATOR = "in"
EQUAL_OPERATOR = "equal"
GT_OPERATOR = "gt"
STARTS_WITH_OPERATOR = "starts_with"
ENDS_WITH_OPERATOR = "ends_with"
CONTAINS_OPERATOR = "contains"
EQUAL_SENSITIVE_OPERATOR = "equal_sensitive"



import pytest

from featureflags.ftypes.string import String


@pytest.mark.parametrize(
    ["arg", "input", "mocked", "method", "expected"],
    [
        ("harness.io", ["harness"], "harness", "starts_with", True),
        ("ci/cd harness", ["harness"], "harness", "ends_with", True),
        ("^harness", ["harness"], "harness", "match", True),
        ("ci/cd harness software", ["harness"], "harness", "contains", True),
        ("harness", ["harness"], "harness", "equal_sensitive", True),
        ("harness", ["harneSS"], "harneSS", "equal_sensitive", False),
        ("harness", ["harneSS"], "harneSS", "equal", True),
        ("harnessb", ["harnessa"], "harnessa", "greater_than", True),
        ("harness", ["harness"], "harness", "greater_than_equal", True),
        ("harnessa", ["harnessb"], "harnessb", "less_than", True),
        ("harness", ["harness"], "harness", "less_than_equal", True),
    ],
)
def test_type_methods(mocker, arg, input, mocked, method, expected):
    m = mocker.patch("featureflags.ftypes.string.get_str_value",
                     return_value=mocked)
    _string = String(value=arg)

    fn = getattr(_string, method)
    got = fn(input)

    assert got == expected
    if mocked:
        m.assert_called_with(input)


def test_in_list_match():
    _string = String(value="Harness")

    got = _string.in_list(["wings", "Harness"])
    expected = True

    assert got == expected


def test_in_list_no_match():
    _string = String(value="Harness")

    got = _string.in_list(["wings", "harness"])
    expected = False

    assert got == expected



import concurrent.futures
import time
from threading import Lock, Thread
from typing import Dict, List, Union

import attr
import httpx

from featureflags.models.metrics_data_metrics_type import \
    MetricsDataMetricsType

from .api.client import AuthenticatedClient
from .api.default.post_metrics import sync_detailed as post_metrics
from .config import Config
from .evaluations.target import Target
from .evaluations.target_attributes import TargetAttributes
from .evaluations.variation import Variation
from .models.key_value import KeyValue
from .models.metrics import Metrics
from .models.metrics_data import MetricsData
from .models.target_data import TargetData
from .models.unset import Unset
from .sdk_logging_codes import (info_metrics_success,
                                info_metrics_target_batch_success,
                                info_metrics_target_exceeded,
                                info_metrics_thread_existed,
                                info_metrics_thread_started,
                                warn_post_metrics_failed,
                                warn_post_metrics_target_batch_failed)
from .util import log

FF_METRIC_TYPE = 'FFMETRICS'
FEATURE_IDENTIFIER_ATTRIBUTE = 'featureIdentifier'
FEATURE_NAME_ATTRIBUTE = 'featureName'
VARIATION_IDENTIFIER_ATTRIBUTE = 'variationIdentifier'
VARIATION_VALUE_ATTRIBUTE = 'variationValue'
TARGET_ATTRIBUTE = 'target'
SDK_VERSION_ATTRIBUTE = 'SDK_VERSION'
SDK_VERSION = '1.0.0'
SDK_TYPE_ATTRIBUTE = 'SDK_TYPE'
SDK_TYPE = 'server'
SDK_LANGUAGE_ATTRIBUTE = 'SDK_LANGUAGE'
SDK_LANGUAGE = 'python'
GLOBAL_TARGET = 'global'


@attr.s(auto_attribs=True)
class AnalyticsEvent(object):
    target: Target
    flag_identifier: str
    variation: Variation
    count: int = 0


@attr.s(auto_attribs=True)
class MetricTargetData(object):
    identifier: str
    name: str
    attributes: Union[Unset, TargetAttributes] = Unset


class AnalyticsService(object):

    def __init__(self, config: Config, client: AuthenticatedClient,
                 environment: str) -> None:
        self._lock = Lock()
        self._config = config
        self._client = client
        self._environment = environment
        self._data: Dict[str, AnalyticsEvent] = {}
        self._target_data_batches: List[Dict[str, MetricTargetData]] = [{}]
        self._max_number_of_batches = 200
        self._max_batch_size = 1000
        self._current_batch_index = 0
        self.max_target_data_exceeded = False

        self._running = False
        self._runner = Thread(target=self._sync)
        self._runner.daemon = True
        self._runner.start()

    def enqueue(self, target: Target, identifier: str,
                variation: Variation):
        event: AnalyticsEvent = AnalyticsEvent(
            target=target,
            flag_identifier=identifier,
            variation=variation
        )

        self._lock.acquire()
        try:
            # Store unique evaluation events. We map a unique evaluation
            # event to its count.
            unique_evaluation_key = self.get_key(event)
            if unique_evaluation_key in self._data:
                self._data[unique_evaluation_key].count += 1
            else:
                event.count = 1
                self._data[unique_evaluation_key] = event

            # Check if we're on our final batch - if we are, and we've
            # exceeded the max batch size just return early.
            if len(self._target_data_batches) >= self._max_number_of_batches:
                if len(self._target_data_batches[
                    self._current_batch_index]) >= \
                        self._max_batch_size:
                    if not self.max_target_data_exceeded:
                        self.max_target_data_exceeded = True
                        info_metrics_target_exceeded()
                    return

            if event.target is not None and not event.target.anonymous:
                unique_target_key = self.get_target_key(event)

                # Store unique targets. If the target already exists
                # in any of the batches, don't continue processing it
                for batch in self._target_data_batches:
                    if unique_target_key in batch:
                        return

                # If we've exceeded the max batch size for the current
                # batch, then create a new batch and start using it.
                if len(self._target_data_batches[
                        self._current_batch_index]) >= self._max_batch_size:
                    self._target_data_batches.append({})
                    self._current_batch_index += 1

                target_name = event.target.name
                # If the target has no name use the identifier
                if not target_name:
                    target_name = event.target.identifier
                self._target_data_batches[
                    self._current_batch_index][unique_target_key] = \
                    MetricTargetData(
                        identifier=event.target.identifier,
                        name=target_name,
                        attributes=event.target.attributes
                )

        finally:
            self._lock.release()

    # Returns a key for unique evaluations events.
    def get_key(self, event: AnalyticsEvent) -> str:
        return '{feature}-{variation}-{value}-{target}'.format(
            feature=event.flag_identifier,
            variation=event.variation.identifier,
            value=event.variation.value,
            target=GLOBAL_TARGET,
        )

    # Returns a key for unique targets. Targets are considered unique
    # if they have different identifiers.
    def get_target_key(self, event: AnalyticsEvent) -> str:
        return event.target.identifier

    def _sync(self) -> None:
        if not self._running:
            info_metrics_thread_started(
                f'{self._config.events_sync_interval}s')
            self._running = True
            while self._running:
                time.sleep(self._config.events_sync_interval)
                self._send_data()

    def _send_data(self) -> None:
        if not bool(self._data):
            log.debug('No metrics data!')
            return
        log.debug('Start sending metrics data')
        self._lock.acquire()
        target_data: List[TargetData] = []
        metrics_data: List[MetricsData] = []
        try:
            for _, event in self._data.items():
                metric_attributes: List[KeyValue] = [
                    KeyValue(FEATURE_IDENTIFIER_ATTRIBUTE,
                             event.flag_identifier),
                    KeyValue(FEATURE_NAME_ATTRIBUTE,
                             event.flag_identifier),
                    KeyValue(VARIATION_IDENTIFIER_ATTRIBUTE,
                             event.variation.identifier),
                    KeyValue(VARIATION_VALUE_ATTRIBUTE, event.variation.value),
                    KeyValue(SDK_TYPE_ATTRIBUTE, SDK_TYPE),
                    KeyValue(SDK_LANGUAGE_ATTRIBUTE, SDK_LANGUAGE),
                    KeyValue(SDK_VERSION_ATTRIBUTE, SDK_VERSION),
                    KeyValue(TARGET_ATTRIBUTE, GLOBAL_TARGET)
                ]

                md = MetricsData(
                    timestamp=int(round(time.time() * 1000)),
                    count=event.count,
                    metrics_type=MetricsDataMetricsType.FFMETRICS,
                    attributes=metric_attributes
                )
                metrics_data.append(md)
            for _, unique_target in self._target_data_batches[0].items():
                self.process_target(target_data, unique_target)

            target_data_batches: List[List[TargetData]] = []
            target_data_batch_index = 0
            # We've already accounted for the first batch, so start processing
            # from the second batch onwards
            for batch in self._target_data_batches[1:]:
                target_data_batches.append([])
                for _, unique_target in batch.items():
                    self.process_target(
                        target_data_batches[target_data_batch_index],
                        unique_target)
                target_data_batch_index += 1

        finally:
            self._data = {}
            self._target_data_batches = [{}]
            self._current_batch_index = 0
            self.max_target_data_exceeded = False
            self._lock.release()

        body: Metrics = Metrics(target_data=target_data,
                                metrics_data=metrics_data)
        try:
            response = post_metrics(client=self._client,
                                    environment=self._environment,
                                    json_body=body)

            log.debug('Metrics server returns: %d', response.status_code)
            if response.status_code >= 400:
                warn_post_metrics_failed(response.status_code)
                return
            if len(target_data_batches) > 0:
                log.info('Sending %s target batches to metrics',
                         len(target_data_batches))
                unique_responses_codes = {}

                # Process batches concurrently
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    futures = []
                    for batch in target_data_batches:
                        # Staggering requests over 0.02 seconds mean that we
                        # will send 200 requests every four seconds, so that
                        # the backend isn't hit too hard.
                        time.sleep(0.02)
                        future = executor.submit(
                            self.process_target_data_batch,
                            batch)
                        futures.append(future)

                    # Wait for all batches to complete
                    concurrent.futures.wait(futures)

                    # Get unique status codes
                    for future in futures:
                        status_code = future.result()
                        if status_code in unique_responses_codes:
                            unique_responses_codes[status_code] += 1
                        else:
                            unique_responses_codes[status_code] = 1

                # Log any error codes
                for unique_code, count in unique_responses_codes.items():
                    if response.status_code >= 400:
                        warn_post_metrics_target_batch_failed(
                            f'{count} batches received code {unique_code}')
                    info_metrics_target_batch_success(
                        f'{count} batches successful')

            info_metrics_success()
        except httpx.RequestError as ex:
            warn_post_metrics_failed(ex)

    def process_target_data_batch(self, target_data_batch):
        batch_request_body: Metrics = Metrics(
            target_data=target_data_batch, metrics_data=[]
        )
        response = post_metrics(
            client=self._client, environment=self._environment,
            json_body=batch_request_body
        )
        return response.status_code

    def process_target(self, target_data, unique_target):
        target_attributes: List[KeyValue] = []
        if not isinstance(unique_target.attributes, Unset):
            for key, value in unique_target.attributes.items():
                # Attribute values need to be sent as string to
                # ff-server so convert all values to strings.
                target_attributes.append(KeyValue(key, str(value)))
        td = TargetData(
            identifier=unique_target.identifier,
            name=unique_target.name,
            attributes=target_attributes
        )
        target_data.append(td)

    def close(self) -> None:
        self._running = False
        if len(self._data) > 0:
            self._send_data()
        info_metrics_thread_existed()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
